@@ -3,13 +3,15 @@ import logging
 
 from asana.error import NotFoundError, InvalidTokenError, ForbiddenError
 from django.apps import apps
+from django.db import IntegrityError, transaction
 from django.core.management.base import BaseCommand, CommandError
-from django.core.urlresolvers import reverse
 from django.utils import six
 
 from djasana.connect import client_connect
-from djasana.models import Attachment, Project, Story, SyncToken, Tag, Task, Team, User, Workspace
+from djasana.models import (
+    Attachment, Project, Story, SyncToken, Tag, Task, Team, User, Webhook, Workspace)
 from djasana.settings import settings
+from djasana.utils import set_webhook
 
 logger = logging.getLogger(__name__)
 
@@ -54,22 +56,23 @@ class Command(BaseCommand):
         )
         parser.add_argument(
             '--nocommit', action='store_false', dest='commit',
-            default=True, help='Will pass commit=False to the backend.'
+            default=True, help='Will not commit changes to the database.'
         )
 
     def handle(self, *args, **options):
-        if options.get('interactive', True):
+        self.commit = not options.get('nocommit')
+        if self.commit and options.get('interactive', True):
             self.stdout.write(
                 'WARNING: This will irreparably synchronize your local database from Asana.')
-            yes_or_no = six.moves.input("Are you sure you wish to continue? [y/N] ")
-            if not yes_or_no.lower().startswith('y'):
+            if not self._confirm():
                 self.stdout.write("No action taken.")
                 return
-        self.commit = not options.get('nocommit')
         self.process_archived = options.get('archive')
         models = self._get_models(options)
         if options.get('verbosity', 0) >= 1:
-            self.stdout.write("Syncronizing data from Asana.")
+            message = "Synchronizing data from Asana."
+            self.stdout.write(message)
+            logger.info(message)
         workspaces = options.get('workspace') or []
         if settings.ASANA_WORKSPACE:
             workspaces.append(settings.ASANA_WORKSPACE)
@@ -81,6 +84,11 @@ class Command(BaseCommand):
         for workspace_id in workspace_ids:
             self._sync_workspace_id(workspace_id, projects, models)
 
+    @staticmethod
+    def _confirm():
+        yes_or_no = six.moves.input("Are you sure you wish to continue? [y/N] ")
+        return yes_or_no.lower().startswith('y')
+
     @staticmethod
     def _get_models(options):
         """Returns a list of models to sync"""
@@ -102,12 +110,14 @@ class Command(BaseCommand):
         return models
 
     def _check_sync_project_id(self, project_id, workspace, models):
+        """If we have a valid sync token for this project sync new events else sync the project"""
         new_sync = False
         try:
             sync_token = SyncToken.objects.get(project_id=project_id)
             try:
                 events = self.client.events.get({'resource': project_id, 'sync': sync_token.sync})
                 self._process_events(project_id, events, workspace, models)
+                self._set_webhook(workspace, project_id)
                 return
             except InvalidTokenError as error:
                 sync_token.sync = error.sync
@@ -117,7 +127,9 @@ class Command(BaseCommand):
                 self.client.events.get({'resource': project_id})
             except InvalidTokenError as error:
                 new_sync = error.sync
-        self._sync_project_id(project_id, workspace, models)
+        is_archived = self._sync_project_id(project_id, workspace, models)
+        if not is_archived:
+            self._set_webhook(workspace, project_id)
         if new_sync:
             SyncToken.objects.create(project_id=project_id, sync=new_sync)
 
@@ -166,6 +178,23 @@ class Command(BaseCommand):
                     ', '.join(bad_list)))
         return project_ids
 
+    def _set_webhook(self, workspace, project_id):
+        """Sets a webhook if the setting is configured and a webhook does not currently exist"""
+        if self.commit and settings.DJASANA_WEBHOOK_URL:
+            webhooks = [webhook for webhook in self.client.webhooks.get_all({
+                'workspace': workspace.remote_id, 'resource': project_id})]
+            if webhooks:
+                # If there is exactly one, and it is active, we are good to go,
+                # else delete them and start a new one.
+                webhooks_ = Webhook.objects.filter(project_id=project_id)
+                if len(webhooks) == webhooks_.count() == 1:
+                    if webhooks[0]['active']:
+                        return
+                for webhook in webhooks:
+                    self.client.webhooks.delete_by_id(webhook['id'])
+                Webhook.objects.filter(id__in=webhooks_.values_list('id', flat=True)[1:]).delete()
+            set_webhook(self.client, project_id)
+
     def _process_events(self, project_id, events, workspace, models):
         project = Project.objects.get(remote_id=project_id)
         for event in events['data']:
@@ -181,10 +210,13 @@ class Command(BaseCommand):
                     self._sync_task(event['resource'], project, models)
             elif event['type'] == 'story':
                 self._sync_story(event['resource'])
-        self.stdout.write(
-            self.style.SUCCESS('Successfully synced project {}.'.format(project.name)))
+        message = 'Successfully synced {0} events for project {1}.'.format(
+            len(events['data']), project.name)
+        self.stdout.write(self.style.SUCCESS(message))
+        logger.info(message)
 
     def _sync_project_id(self, project_id, workspace, models):
+        """Sync this project by polling it. Returns boolean 'is archived?'"""
         project_dict = self.client.projects.find_by_id(project_id)
         logger.debug('Sync project %s', project_dict['name'])
         logger.debug(project_dict)
@@ -210,15 +242,6 @@ class Command(BaseCommand):
             follower_ids = [follower['id'] for follower in followers_dict]
             followers = User.objects.filter(id__in=follower_ids)
             project.followers.set(followers)
-            if settings.DJASANA_WEBHOOK_URL:
-                target = '{}{}'.format(
-                    settings.DJASANA_WEBHOOK_URL,
-                    reverse('djasana_webhook', kwargs={'remote_id': project_id}))
-                self.client.webhooks.create({
-                    'resource': project_id,
-                    'target': target,
-                })
-                logger.debug('Setting webhook at {}', target)
         else:
             project = None
 
@@ -227,8 +250,10 @@ class Command(BaseCommand):
                 self._sync_task(task, project, models)
 
         if project:
-            self.stdout.write(
-                self.style.SUCCESS('Successfully synced project {}.'.format(project.name)))
+            message = 'Successfully synced project {}.'.format(project.name)
+            self.stdout.write(self.style.SUCCESS(message))
+            logger.info(message)
+        return project_dict['archived']
 
     def _sync_story(self, story):
         try:
@@ -243,8 +268,12 @@ class Command(BaseCommand):
                 remote_id=story_dict['created_by']['id'],
                 defaults={'name': story_dict['created_by']['name']})[0]
             story_dict['created_by'] = user
-        story_dict.pop('hearts', None)
-        story_dict['target'] = story_dict['target']['id']
+        if story_dict['target']:
+            story_dict['target'] = story_dict['target']['id']
+        for key in ('hearts', 'liked', 'likes', 'num_likes'):
+            story_dict.pop(key, None)
+        if 'text' in story_dict:
+            story_dict['text'] = story_dict['text'][:1024]  # Truncate text if too long
         Story.objects.get_or_create(remote_id=remote_id, defaults=story_dict)
 
     def _sync_tag(self, tag):
@@ -256,10 +285,15 @@ class Command(BaseCommand):
                 remote_id=remote_id,
                 defaults=tag_dict)
 
-    def _sync_task(self, task, project, models):
+    def _sync_task(self, task, project, models, skip_subtasks=False):
+        """Sync this task and parent its subtasks
+
+        For parents and subtasks, this method is called recursively, so skip_subtasks True is
+        passed when syncing a parent task from a subtask.
+        """
         try:
             task_dict = self.client.tasks.find_by_id(task['id'])
-        except ForbiddenError:
+        except (ForbiddenError, NotFoundError):
             try:
                 Task.objects.get(remote_id=task['id']).delete()
             except Task.DoesNotExist:
@@ -275,16 +309,25 @@ class Command(BaseCommand):
                     remote_id=task_dict['assignee']['id'],
                     defaults={'name': task_dict['assignee']['name']})[0]
                 task_dict['assignee'] = user
-            task_dict.pop('hearts', None)
-            task_dict.pop('memberships')
-            task_dict.pop('projects')
-            task_dict.pop('workspace')
-            if task_dict['parent']:
-                self._sync_task(task_dict['parent'], project, models)
+            for key in (
+                    'hearts', 'liked', 'likes', 'num_likes',
+                    'memberships', 'projects', 'workspace'):
+                task_dict.pop(key, None)
+            parent = task_dict.pop('parent', None)
+            if parent:
+                # If this is a task we already know about, assume it was just synced.
+                parent_id = parent['id']
+                if not Task.objects.filter(remote_id=parent_id).exists():
+                    self._sync_task(parent, project, models, skip_subtasks=True)
+                task_dict['parent_id'] = parent_id
             followers_dict = task_dict.pop('followers')
             tags_dict = task_dict.pop('tags')
             task_ = Task.objects.update_or_create(
                 remote_id=remote_id, defaults=task_dict)[0]
+            if not skip_subtasks:
+                subtasks = self.client.tasks.subtasks(task['id'])
+                for subtask in subtasks:
+                    self._sync_task(subtask, project, models)
             follower_ids = [follower['id'] for follower in followers_dict]
             followers = User.objects.filter(id__in=follower_ids)
             task_.followers.set(followers)
@@ -326,10 +369,11 @@ class Command(BaseCommand):
             user_dict.pop('workspaces')
             if user_dict['photo']:
                 user_dict['photo'] = user_dict['photo']['image_128x128']
-            user = User.objects.get_or_create(
+            user = User.objects.update_or_create(
                 remote_id=remote_id,
                 defaults=user_dict)[0]
-            user.workspaces.add(workspace)
+            if workspace:
+                user.workspaces.add(workspace)
 
     def _sync_workspace_id(self, workspace_id, projects, models):
         workspace_dict = self.client.workspaces.find_by_id(workspace_id)
@@ -363,5 +407,6 @@ class Command(BaseCommand):
                 self._check_sync_project_id(project_id, workspace, models)
 
         if workspace:
-            self.stdout.write(
-                self.style.SUCCESS('Successfully synced workspace {}.'.format(workspace.name)))
+            message = 'Successfully synced workspace {}.'.format(workspace.name)
+            self.stdout.write(self.style.SUCCESS(message))
+            logger.info(message)
